% 
% Solution of exercise 1 for the lab session on matrix equations and
% hierarchical matrices. 
%

a = -1;
b = 1;

% Some parameters:
n = 2^10; % Number of discretization points on the edge of the grid. 

% Problem statement: we would like to solve a 2D PDE on [0, 1]^2, given by 
%    \Delta u + f = 0, 
% subject to Dirichlet zero boundary conditions. We assume f is given. 

f = @(x,y) exp(- x.^2 - y.^2);

% We note that f(x) = g1(x) g2(y), where g1(x) and g2(y) are defined as
% follows:
g1 = @(x) exp(-x.^2);
g2 = @(y) exp(-y.^2);

% Construct the 1D discretization of the Laplacian. 
h = (b - a) ./ (n - 1);
x = linspace(a, b, n);

% The system has size (n-2) x (n-2) because we have imposed the Dirichlet
% boundary conditions.
A = h^(-2) * spdiags(ones(n-2, 1) * [1 -2 1], -1:1, n - 2, n - 2);

% We can now solve the matrix equation AX + XA + F = 0, using the
% factorization F = UF * VF';
UF = g1(x(2:end-1)).'; VF = g2(x(2:end-1)).';

% inf is the maximum number of iterations, 1e-6 the tolerance, true enables
% the debug. The last parameter is optional, but switches the stopping
% criterion to the Frobenius norm, which in practice correspond to a scaled
% version of the (discrete) L^2 norm. 
[UX, VX] = ek_sylv(A, A, UF, VF, inf, 1e-6, true, 'fro');

% Plot the solution, if it's not too large, otherwise we subsample it to k
% points where k is small enough
if n <= 1024
	mesh(x, x, blkdiag(0, UX * VX', 0));
else	
	k = 256;
	indices = ceil( (1:k) ./ k * size(UX, 1) );

	% Include boundary conditions for the indices of x. 
	allindices = [ 1, indices, n ];

	% Add the zero Dirichlet boundary conditions
	X = blkdiag(0, UX(indices, :) * VX(indices, :)', 0);
	
	% Plot the subsampled mesh
	mesh(x(allindices), x(allindices), X);
end