% 
% Solution of exercise 5 for the lab session on matrix equations and
% hierarchical matrices. 
%
% This is the exercise 1 adapted to work with HODLR matrices. The same
% could be obtain with essentially no changes with all the other exercises.

% Some parameters:
n = 2^10; % Number of discretization points on the edge of the grid. 

% Problem statement: we would like to solve a 2D PDE on [0, 1]^2, given by 
%    \Delta u + f = 0, 
% subject to Dirichlet zero boundary conditions. We assume f is given. 

f = @(x,y) log(1 + abs(x - y));

% Construct the 1D discretization of the Laplacian. 
h = 1 ./ (n - 1);
x = linspace(0, 1, n);
y = x;

% The system has size (n-2) x (n-2) because we have imposed the Dirichlet
% boundary conditions.
A = h^(-2) * spdiags(ones(n-2, 1) * [1 -2 1], -1:1, n - 2, n - 2);

% For this to work, we need to work with HODLR representations of A as well
A = hodlr('banded', A);

% Build an hodlr representation of the right hand side F
F = hodlr('handle', @(i,j) f(x(1 + j), y(1 + i).'), n - 2, n - 2);

% Solve the Lyapunov equation
X = lyap(A, A, F);

% Plot the solution, if it's not too large, otherwise we subsample it to k
% points where k is small enough
if n <= 256
	mesh(x, x, blkdiag(0, full(X), 0));
else	
	k = 256;
	indices = ceil( (1:k) ./ k * size(X, 1) );

	% Include boundary conditions for the indices of x. 
	allindices = [ 1, indices, n ];

	% Add the zero Dirichlet boundary conditions
	X = blkdiag(0, full(X(indices,indices)), 0);
	
	% Plot the subsampled mesh
	mesh(x(allindices), x(allindices), X);
end